#include <avr/io.h>
#include <stdio.h>
#include <math.h>

#include <util/delay.h>
#include <stdlib.h>

#define BAUDRATE 9600						//set desired baud rate
#define MYUBRR (F_CPU/8/BAUDRATE-1)			//calculate UBRR value

void USART_Init(unsigned int ubrr){
	UBRR0H = (unsigned char)(ubrr>>8);  	//low byte
	UBRR0L = (unsigned char)ubrr; 			//high byte
	UCSR0B |= (1<<RXEN0)|(1<<TXEN0);		//Enable Transmitter and Receiver	
	UCSR0A |= (1<<U2X0);
}

void USART_vSendByte_Pc(uint8_t u8Data, FILE *stream){
    while((UCSR0A &(1<<UDRE0)) == 0);		// Wait if a byte is being transmitted
	UDR0 = u8Data;  						// Transmit data
}
static FILE mystdout = FDEV_SETUP_STREAM(USART_vSendByte_Pc, NULL, _FDEV_SETUP_WRITE);


void matrisemul( double *CC, double *AA,  double *BB,int sizeAm, int sizeAn, int sizeBm,int sizeBn)
{
    double sum;
    for (int i = 0; i<sizeAn;i++){
        for(int j = 0; j<sizeBn; j++){
            sum = 0;
            for(int k = 0; k<sizeBm;k++){
            sum = sum + AA[i*sizeAn+k]*BB[k*sizeBn+j];

            }
            CC[i*sizeBn+j] = sum;

        }
    }
}

void PWM_init(){
	// Waveform generation mode: PWM, Phase and Frequency Correct
    TCCR1A |= (0<<WGM11)|(0<<WGM10);
	TCCR1B |= (1<<WGM13)|(0<<WGM12);

	// No prescaling 
	TCCR1B |= (0<CS12)|(0<<CS11)|(1<<CS10);

	// Clear on compare match when up-counting. Set on compare match when downcounting.
	TCCR1A |= (1<<COM1A1)|(0<<COM1A0);
	TCCR1A |= (1<<COM1B1)|(0<<COM1B0);
	TCCR1A |= (1<<COM1C1)|(0<<COM1C0);

	// No filtering
	TCCR1B |= (0<<ICNC1)|(0<<ICES1);

	// Resolution
	ICR1 = 0x0400 - 1;
	
 	DDRB |= (1<<DDB0)|(1<<DDB4);
}

int main(void){
	PWM_init();
	USART_Init(MYUBRR);
	stdout = &mystdout;	

	// Initialize variables variables
    double lat, w_B_IB[3], quaternions[4], eps[3], eta;
    double Rc;
	double w_o;
	double V;
	//double Rx, ix_max, Nx, Ax, ix;
	//double Ry, iy_max, Ny, Ay, iy;
	double Rz, iz_max, Nz, Az, iz;
	double pi;
    double R11, R12, R13, R21, R22, R23, R31, R32, R33;
    double R_O_B[9], R_B_O[9], w_B_OB[3];
	double detumblingGain;
    double Bn, Br;
    double B_O[3], B_B[3], B_B_normSq;
    double m_B[3];
    
	// Orbit parameters
    Rc = 6971200;
	w_o = sqrt((6.67428*pow(10,-11)*5.972*pow(10,24))/pow(Rc,3));

	// Coil parameters
	V = 5;
	Rz = 100; iz_max = V/Rz, Nz = 784; Az = 0.08*0.08;
	// Define y-coil values
	// Define z-coil values

	// PI definition
	pi = 3.1415926535;

	// Input variables
    lat = 0;
	w_B_IB[0] = 0;
    w_B_IB[1] = -w_o-0.2;
    w_B_IB[2] = 0;
	quaternions[0] = 0;
    quaternions[1] = 0;
    quaternions[2] = 0;
    quaternions[3] = 1;
    eps[0] = quaternions[0]; eps[1] = quaternions[1]; eps[2] = quaternions[2]; eta = quaternions[3];

	// R_O_B = eye(3) + 2*eta*S_eps + 2*S_eps^2;
	R11 = 1-2*eps[1]*eps[1]-2*eps[2]*eps[2];
	R12 = -2*eta*eps[2]+2*eps[0]*eps[1];
	R13 = 2*eta*eps[1]+2*eps[0]*eps[2];
	R21 = 2*eta*eps[2]+2*eps[0]*eps[1];
	R22 = 1-2*eps[0]*eps[0]-2*eps[2]*eps[2];
	R23 = -2*eta*eps[0]+2*eps[1]*eps[2];
	R31 = 2*eta*eps[1]+2*eps[0]*eps[2];
	R32 = 2*eta*eps[0]+2*eps[1]*eps[2];
	R33 = 1-2*eps[0]*eps[0]-2*eps[1]*eps[1];
    R_O_B[0] = R11; R_O_B[1] = R12; R_O_B[2] = R13;
    R_O_B[3] = R21; R_O_B[4] = R22; R_O_B[5] = R23;
    R_O_B[6] = R31; R_O_B[7] = R32; R_O_B[8] = R33;
	// R_B_O = R_O_B';
    R_B_O[0] = R11; R_B_O[1] = R21; R_B_O[2] = R31;
    R_B_O[3] = R12; R_B_O[4] = R22; R_B_O[5] = R32;
    R_B_O[6] = R13; R_B_O[7] = R23; R_B_O[8] = R33;

	// w_O_IO = [0;-w_o;0];
	// w_B_OB = w_B_IB-R_B_O*w_O_IO;
	w_B_OB[0] = w_B_IB[0]+R21*w_o;
    w_B_OB[1] = w_B_IB[1]+R22*w_o;
    w_B_OB[2] = w_B_IB[2]+R23*w_o;

	// Detumbling controller
	// [tau_m,w] = detumblingController1(P,w_B_OB,lat)
	// NEED: To implement IGRF for B_O
	B_O[0] = 2.3489*pow(10,-5);
    B_O[1] = 0;
    B_O[2] = 0;
	// B_B = R_B_O*B_O;
	matrisemul(B_B,R_B_O,B_O,3,3,3,1);

	// d = 4e-5
	// m_B = -d*cross(B_B,w_B_OB);
	B_B_normSq = B_B[0]*B_B[0]+B_B[1]*B_B[1]+B_B[2]*B_B[2];
	B_B_normSq = B_B_normSq*B_B_normSq;
	detumblingGain = 4*pow(10,-5);
	m_B[0] = (detumblingGain/B_B_normSq)*B_B[2]*w_B_OB[1]-pow(10,4)*B_B[1]*w_B_OB[2];
    m_B[1] = -(detumblingGain/B_B_normSq)*B_B[2]*w_B_OB[0]+pow(10,4)*B_B[0]*w_B_OB[2];
    m_B[2] = 2*(detumblingGain/B_B_normSq)*B_B[1]*w_B_OB[0]-2*pow(10,4)*B_B[0]*w_B_OB[1];
	
	// [m_B,w] = momentumScaling(P,m_B);
	// ix = m_B[0]/(Nx*Ax);
	// iy = m_B[1]/(Ny*Ay);
	iz = m_B[2]/(Nz*Az);
	if (iz > iz_max){
		iz = iz_max;
		}
	// NEED: More complex scaling structure needed for multiple coils, in order to preserve tau_m's direction


	// PB5 PB6 and PB7 are used as PWM outputs. Input to ENABLE on H-bridge
	// PB0 PB1 PB2 PB3 and PB4 can be used as forward/reverse operation outputs. Inputs to PHASE on H-bridge.

	while(1){
		OCR1A = (0x0400-1)*(iz/iz_max);
		if (iz < 0)
			PORTB |= (1<<DDB0);
		else
			PORTB |= (0<<DDB0);

	// printf("%ld\n",(long)(R22*10));
	// printf("%d\n",(int)(R22*1000));
	// printf("%ld\n",(TCNT1 & 11111111));
	}
}


