#include <avr/io.h>
#include <stdio.h>
#include <math.h>

#include <util/delay.h>
#include <stdlib.h>

#include "IGRFcoeffs.h"

#define BAUDRATE 9600						//set desired baud rate
#define MYUBRR (F_CPU/8/BAUDRATE-1)			//calculate UBRR value

void USART_Init(unsigned int ubrr){
	UBRR0H = (unsigned char)(ubrr>>8);  	//low byte
	UBRR0L = (unsigned char)ubrr; 			//high byte
	UCSR0B |= (1<<RXEN0)|(1<<TXEN0);		//Enable Transmitter and Receiver	
	UCSR0A |= (1<<U2X0);
}

void USART_vSendByte_Pc(uint8_t u8Data, FILE *stream){
    while((UCSR0A &(1<<UDRE0)) == 0);		// Wait if a byte is being transmitted
	UDR0 = u8Data;  						// Transmit data
}
static FILE mystdout = FDEV_SETUP_STREAM(USART_vSendByte_Pc, NULL, _FDEV_SETUP_WRITE);

static const double pi, Re, Rc;
// Coil and power supply constants
static const double V;
static const double Rz, iz_max, Nz, Az;


void PWM_init(){
	// Waveform generation mode: PWM, Phase and Frequency Correct
    TCCR1A |= (0<<WGM11)|(0<<WGM10);
	TCCR1B |= (1<<WGM13)|(0<<WGM12);

	// No prescaling 
	TCCR1B |= (0<CS12)|(0<<CS11)|(1<<CS10);

	// Clear on compare match when up-counting. Set on compare match when downcounting.
	TCCR1A |= (1<<COM1A1)|(1<<COM1A0);
	TCCR1A |= (1<<COM1B1)|(1<<COM1B0);
	TCCR1A |= (1<<COM1C1)|(1<<COM1C0);

	// No filtering
	TCCR1B |= (0<<ICNC1)|(0<<ICES1);

	// Resolution
	ICR1 = 0x0400 - 1;

 	// DDRB |= (1<<DDB0)|(1<<DDB4);
	DDRB = 0xFF;
}

void IGRF(double *out, unsigned short int n, unsigned short int m, double lat, double lon){
// IGRF START
	int i, j;
	double dP[(n+1)*(m+1)+1];
	for(int k=0;k<=(n+1)*(m+1);++k)
		dP[k] = 0;
	double P[(n+1)*(m+1)+1];
	for(int k=0;k<=(n+1)*(m+1);++k)
		P[k] = 1;
	double S[(n+1)*(m+1)+1];
	for(int k=0;k<=(n+1)*(m+1);++k)
		S[k] = 1;
	double deltaFun, K;
	double mt, tol;
	double theta, phi;
	double Bt2, Bp2, Br2;
	double Bt1, Bp1, Br1;

	Bt2 = 0; Bp2 = 0; Br2 = 0;

	i = n;
	j = m;
	tol = pow(10,-8);

	mt = ((int)(lat*180/pi) % (360));
	mt = mt*pi/180;

	if (mt>((pi/2)-tol) && mt<((pi/2)+tol)){
		lat = 90+tol;
	}
	else if(mt>((3*pi/2)-tol) && mt<((3*pi/2)+tol)){
		lat = 270+tol;
	}

	theta = (pi/2)-lat;
	phi = lon;

	for (n=0;n<=i;++n){;
		for (m=0;m<=j;++m){
			if (n>0) {
				if (m==0){
					S[n*(j+1)+m] = S[(n-1)*(j+1)]*(2*n-1)/n;
	
				}
				else{
					if (m==1)
						deltaFun = 1;
					else
						deltaFun = 0;
					S[n*(j+1)+m] = S[n*(j+1)+(m-1)]*sqrt(((n-m+1)*(deltaFun+1))/(n+m));
				}
			}
		}
	}
	
	for (n=0;n<=i;++n){
		for (m=0;m<=j;++m){
			if (n==1){
				dP[n*(j+1)+m] = cos(theta)*dP[(n-1)*(j+1)+m]-sin(theta)*P[(n-1)*(j+1)+m];
				P[n*(j+1)+m] = cos(theta)*P[(n-1)*(j+1)+m];
				
			}
			else if (n>1){
				K = (pow(n-1,2)-pow(m,2))/((2*n-1)*(2*n-3));
				dP[n*(j+1)+m] = cos(theta)*dP[(n-1)*(j+1)+m]-sin(theta)*P[(n-1)*(j+1)+m]-K*dP[(n-2)*(j+1)+m];
				P[n*(j+1)+m] = cos(theta)*P[(n-1)*(j+1)+m]-K*P[(n-2)*(j+1)+m];
			}
			if ((n==m) && (n>0)){
				P[n*(j+1)+m] = sin(theta)*P[(n-1)*(j+1)+(n-1)];
				dP[n*(j+1)+m] = sin(theta)*dP[(n-1)*(j+1)+(n-1)]+cos(theta)*P[(n-1)*(j+1)+(n-1)];
			}
		}
	}

	for (n=1;n<=i;++n){
		Bt1 = 0;
		Bp1 = 0;
		Br1 = 0;
		for (m=0;m<=j;++m){
			Bt1 = Bt1 + (S[n*(j+1)+m]*g_data[n*(13+1)+m]*cos(m*phi)+S[n*(j+1)+m]*h_data[n*(13+1)+m]*sin(m*phi))*dP[n*(j+1)+m];
			Bp1 = Bp1 + (m*S[n*(j+1)+m]*h_data[n*(13+1)+m]*cos(m*phi)-m*S[n*(j+1)+m]*g_data[n*(13+1)+m]*sin(m*phi))*P[n*(j+1)+m];
			Br1 = Br1 + (S[n*(j+1)+m]*g_data[n*(13+1)+m]*cos(m*phi)+S[n*(j+1)+m]*h_data[n*(13+1)+m]*sin(m*phi))*P[n*(j+1)+m];
		}
		Bt2 = Bt2 + (pow(Re/Rc,n+2)*Bt1);
		Bp2 = Bp2 + (pow(Re/Rc,n+2)*Bp1);
		Br2 = Br2 + (n+1)*(pow(Re/Rc,n+2)*Br1);
	}

	// In Orbit frame, in Cartesian coordinates
	double dummyEps;
	dummyEps = 0;
	out[0] = (Bt2*cos(dummyEps)-Br2*sin(dummyEps))*pow(10,-9);
	out[1] = (-Bp2/sin(theta))*pow(10,-9);
	out[2] = (-Bt2*sin(dummyEps)-Br2*cos(dummyEps))*pow(10,-9);

	/* TESTING
    printf("B_O[0]=%d\n",(int)(out[0]*10000000));
	printf("B_O[1]=%d\n",(int)(out[1]*10000000));
	printf("B_O[2]=%d\n",(int)(out[2]*10000000));
	*/
	
}

double currentScaling(double *m_B){
	double iz;

	iz = m_B[2]/(Nz*Az);
	printf("iz=%d, m_B[2]=%d\n",(int)(iz*1000),(int)(m_B[2]*100));
	if (abs(iz) > iz_max){
		if (iz > 0)
			return iz_max;
		else
			return -iz_max;
	}
	else {
		return iz;
	}
	// NEED: More complex scaling structure needed for multiple coils, in order to preserve tau_m's direction
}

int main(void){
	PWM_init();
	USART_Init(MYUBRR);
	stdout = &mystdout;	

	// Initialize variables variables
    double lat, lon , w_B_IB[3], quaternions[4], eps[3], eta;
	double w_o;	
    double R11, R12, R13, R21, R22, R23, R31, R32, R33;
    double R_O_B[9], R_B_O[9], w_B_OB[3];
	double detumblingGain;
    double B_O[3], B_B[3], B_B_normSq;
    double m_B[3], iz;
    
	// Orbit parameters
	w_o = sqrt((6.67428*pow(10,-11)*5.972*pow(10,24))/pow(Rc,3));

	// Input variables
    lat = 0; lon = 0;
	w_B_IB[0] = 0;
    w_B_IB[1] = -w_o-0.2;
    w_B_IB[2] = 0;
	quaternions[0] = 0;
    quaternions[1] = 0;
    quaternions[2] = 0;
    quaternions[3] = 1;
    eps[0] = quaternions[0]; eps[1] = quaternions[1]; eps[2] = quaternions[2]; eta = quaternions[3];

	// MATLAB CMD: R_O_B = eye(3) + 2*eta*S_eps + 2*S_eps^2;
	R11 = 1-2*eps[1]*eps[1]-2*eps[2]*eps[2];
	R12 = -2*eta*eps[2]+2*eps[0]*eps[1];
	R13 = 2*eta*eps[1]+2*eps[0]*eps[2];
	R21 = 2*eta*eps[2]+2*eps[0]*eps[1];
	R22 = 1-2*eps[0]*eps[0]-2*eps[2]*eps[2];
	R23 = -2*eta*eps[0]+2*eps[1]*eps[2];
	R31 = 2*eta*eps[1]+2*eps[0]*eps[2];
	R32 = 2*eta*eps[0]+2*eps[1]*eps[2];
	R33 = 1-2*eps[0]*eps[0]-2*eps[1]*eps[1];
    R_O_B[0] = R11; R_O_B[1] = R12; R_O_B[2] = R13;
    R_O_B[3] = R21; R_O_B[4] = R22; R_O_B[5] = R23;
    R_O_B[6] = R31; R_O_B[7] = R32; R_O_B[8] = R33;

	// MATLAB CMD: R_B_O = R_O_B';
    R_B_O[0] = R11; R_B_O[1] = R21; R_B_O[2] = R31;
    R_B_O[3] = R12; R_B_O[4] = R22; R_B_O[5] = R32;
    R_B_O[6] = R13; R_B_O[7] = R23; R_B_O[8] = R33;

	// MATLAB CMD: w_B_OB = w_B_IB-R_B_O*w_O_IO;
	w_B_OB[0] = w_B_IB[0]+R21*w_o;
    w_B_OB[1] = w_B_IB[1]+R22*w_o;
    w_B_OB[2] = w_B_IB[2]+R23*w_o;

	// ##START DISSIPATIVE DETUMBLING CONTROLLER SECTION##
	// Calculate geomagnetic field from IGRF model
	IGRF(B_O,13,13,lat,lon);

	// MATLAB CMD: B_B = R_B_O*B_O;
	B_B[0] = R_B_O[0]*B_O[0]+R_B_O[1]*B_O[1]+R_B_O[2]*B_O[2];
	B_B[1] = R_B_O[3]*B_O[0]+R_B_O[4]*B_O[1]+R_B_O[5]*B_O[2];
	B_B[2] = R_B_O[6]*B_O[0]+R_B_O[7]*B_O[1]+R_B_O[8]*B_O[2];

	/* TESTING
    printf("B_B[0]=%d\n",(int)(B_B[0]*10000000));
	printf("B_B[1]=%d\n",(int)(B_B[1]*10000000));
	printf("B_B[2]=%d\n",(int)(B_B[2]*10000000));
	*/
	
	// MATLAB CMD: norm(B_B,2)^2;
	B_B_normSq = B_B[0]*B_B[0]+B_B[1]*B_B[1]+B_B[2]*B_B[2];
	// Controller gain: d = 4e-5
	detumblingGain = 4*pow(10,-5);
	// MATLAB CMD: m_B = -(d/norm(B_B,2)^2)*cross(B_B,w_B_OB);
	m_B[0] = (detumblingGain/B_B_normSq)*(B_B[2]*w_B_OB[1]-B_B[1]*w_B_OB[2]);
    m_B[1] = (detumblingGain/B_B_normSq)*(-B_B[2]*w_B_OB[0]+B_B[0]*w_B_OB[2]);
    m_B[2] = (detumblingGain/B_B_normSq)*(B_B[1]*w_B_OB[0]-B_B[0]*w_B_OB[1]);

	// MATLAB CMD: [m_B,w] = currentScaling(P,m_B);
	iz = currentScaling(m_B);
	// ##END DISSIPATIVE DETUMBLING CONTROLLER SECTION##

	/* TESTING
	printf("w_B_OB{1,2}=%d,%d\n",(int)(w_B_OB[0]*1000),(int)(w_B_OB[1]*1000));
	printf("B_B{1,2}=%d,%d\n",(int)(B_B[0]*10000000),(int)(B_B[1]*10000000));
	printf("iz=%d, m_B[2]=%d\n",(int)(iz*1000),(int)(m_B[2]*100));
	*/

	// PB5 PB6 and PB7 are used as PWM outputs. Input to ENABLE on H-bridge
	// PB0 PB1 PB2 PB3 and PB4 can be used as forward/reverse operation outputs. Inputs to PHASE on H-bridge.

	while(1){
		// Setting compare PWM register by the relation of desired current to the maximum current
		OCR1A = (0x0400-1)*(iz/iz_max);
		//OCR1A = 0x03FF;
		if (iz > 0)
			PORTB |= (1<<DDB0);
		else
			PORTB |= (0<<DDB0);
	}

}


