addpath('myIGRF');

w_o = P.w_o;
Rc = P.Rc;

my_m = 1e17/(4*pi);
B0 = my_m/Rc^3;

B_O = [];
lat = 0:0.1:2*pi;
for i=0:0.1:2*pi
    Bn = B0*cos(i);       % x-axis in orbital frame, parallel to the velocity vector
    Br = 2*B0*sin(i);     % z-axis in orbital frame, toward Earth center
    B_O = [B_O;Bn, 0, Br];
end

subplot(3,1,1);
plot(lat,B_O(:,1),'k');
hold on
plot(lat,B_O(:,2),':k');
plot(lat,B_O(:,3),'--k');
title('Dipole field model in O-frame');
ylabel('[T] Magnitude');
xlabel('[rad] Latitude');
legend('x_o','y_o','z_o');
axis([0 2*pi -5e-5, 5e-5]);

igrfField = [];
for i=0:0.1:2*pi
    field = Bfunk(13,13,i,0*pi/180);
    igrfField = [igrfField;field'];
end

subplot(3,1,2);
plot(lat,igrfField(:,1),'k');
hold on
plot(lat,igrfField(:,2),':k');
plot(lat,igrfField(:,3),'--k');
title('IGRF-11 model in O-frame Lon=0');
ylabel('[T] Magnitude');
xlabel('[rad] Latitude');
legend('x_o','y_o','z_o');
axis([0 2*pi -5e-5, 5e-5]);

igrfField = [];
for i=0:0.1:2*pi
    field = Bfunk(13,13,i,90*pi/180);
    igrfField = [igrfField;field'];
end

subplot(3,1,3);
plot(lat,igrfField(:,1),'k');
hold on
plot(lat,igrfField(:,2),':k');
plot(lat,igrfField(:,3),'--k');
title('IGRF-11 model in O-frame Lon=\pi/2');
ylabel('[T] Magnitude');
xlabel('[rad] Latitude');
legend('x_o','y_o','z_o');
axis([0 2*pi -5e-5, 5e-5]);